from pathlib import Path
from typing import Union

from pydantic import BaseModel

from hyfi.hydra import Composer
from hyfi.module import ModuleConfig
from hyfi.path.batch import BatchPathConfig
from hyfi.project import ProjectConfig
from hyfi.utils.lib import ensure_import_module
from hyfi.utils.logging import getLogger

logger = getLogger(__name__)


class TaskConfig(BaseModel):
    config_name: str = "__init__"
    config_group: str = "task"
    task_name: str = "demo-task"
    task_root: str = "tmp/task"
    autoload: bool = False
    version: str = "0.0.0"
    module: ModuleConfig = None  # type: ignore
    path: BatchPathConfig = None  # type: ignore
    project: ProjectConfig = None  # type: ignore

    class Config:
        arbitrary_types_allowed = True
        extra = "allow"
        validate_assignment = False
        exclude = {
            "__data__",
            "path",
            "module",
            "project",
        }
        include = {}
        underscore_attrs_are_private = True
        property_set_methods = {
            "task_name": "set_task_name",
            "task_root": "set_task_root",
        }

    def __init__(
        self,
        config_name: str = "__init__",
        config_group: str = "task",
        **data,
    ):
        super().__init__(**data)
        self.initialize_configs(
            config_name=config_name,
            config_group=config_group,
            **data,
        )

    def __setattr__(self, key, val):
        super().__setattr__(key, val)
        if method := self.__config__.property_set_methods.get(key):  # type: ignore
            getattr(self, method)(val)

    def set_task_root(self, val: Union[str, Path]):
        if not self.task_root or self.task_root != val:
            self.initialize_configs(task_root=val)

    def set_task_name(self, val):
        if not self.task_name or self.task_name != val:
            self.initialize_configs(task_name=val)

    def initialize_configs(
        self,
        config_name: str = "__init__",
        config_group: str = "task",
        **data,
    ):
        # Initialize the config with the given config_name.
        data = Composer(
            config_group=f"{config_group}={config_name}",
            config_data=data,
        ).config_as_dict
        self.__dict__.update(data)
        if "module" in data:
            self.module = ModuleConfig(**data["module"])
        if "path" in data:
            self.path = BatchPathConfig(**data["path"])
        if "project" in data:
            self.project = ProjectConfig(**data["project"])

    @property
    def config(self):
        return self.dict()

    @property
    def root_dir(self) -> Path:
        return self.path.root_dir

    @property
    def output_dir(self) -> Path:
        return self.path.output_dir

    @property
    def project_name(self):
        return self.project.project_name

    @property
    def project_dir(self) -> Path:
        return Path(self.project.project_root)

    @property
    def workspace_dir(self) -> Path:
        return Path(self.project.project_workspace_dir)

    @property
    def model_dir(self) -> Path:
        return self.path.model_dir

    @property
    def log_dir(self) -> Path:
        return self.project.path.log_dir

    @property
    def cache_dir(self) -> Path:
        return self.project.path.cache_dir

    @property
    def library_dir(self) -> Path:
        return self.path.library_dir

    @property
    def verbose(self) -> bool:
        return bool(self.project.verbose)

    def print_config(self):
        Composer.print(self.config)

    def load_modules(self):
        """Load the modules"""
        if not self.module.modules:
            logger.info("No modules to load")
            return
        library_dir = self.library_dir
        for module in self.module.modules:
            name = module.name
            libname = module.libname
            liburi = module.liburi
            specname = module.specname
            libpath = library_dir / libname
            syspath = module.get("syspath")
            if syspath is not None:
                syspath = library_dir / syspath
            ensure_import_module(name, libpath, liburi, specname, syspath)

    def reset(self, objects=None, release_gpu_memory=True):
        """Reset the memory cache"""
        if isinstance(objects, list):
            for obj in objects:
                del obj
        if release_gpu_memory:
            from hyfi.utils.gpu import GPUMon

            GPUMon.release_gpu_memory()



from pathlib import Path
from typing import Any

from hyfi.hydra import Composer
from hyfi.path.task import TaskPathConfig
from hyfi.utils.logging import getLogger

logger = getLogger(__name__)


class BatchPathConfig(TaskPathConfig):
    config_name: str = "__batch__"

    batch_outputs: str = ""

    class Config:
        extra = "ignore"

    def __init__(
        self,
        config_name: str = "__batch__",
        config_group: str = "path",
        **data: Any,
    ):
        """
        Initialize the batch. This is the method you call when you want to initialize the batch from a config

        Args:
            config_name: The name of the config you want to use
            data: The data you want to initilize the
        """
        # Initialize the config with the given config_name.
        super().__init__(**data)
        self.initialize_configs(
            config_name=config_name,
            config_group=config_group,
            **data,
        )

    def initialize_configs(
        self,
        config_name: str = "__batch__",
        config_group: str = "path",
        **data,
    ):
        # Initialize the config with the given config_name.
        data = Composer(
            config_group=f"{config_group}={config_name}",
            config_data=data,
        ).config_as_dict
        self.__dict__.update(data)

    @property
    def batch_dir(self) -> Path:
        """
        The directory where the batch is stored. It is used to determine where the results are stored for a batch of data to be processed.


        Returns:
                The directory where the batch output is stored for a batch of data to be processed ( relative to the task output directory )
        """
        self.batch_outputs = (
            self.batch_outputs or (self.output_dir / "batch-outputs").as_posix()
        )
        return Path(self.batch_outputs).absolute()



"""
A class to apply a pipe to a dataframe or a dictionary of dataframes.
"""
from tqdm.auto import tqdm

from hyfi.hydra import SpecialKeys
from hyfi.hydra.main import XC
from hyfi.joblib.batch import batcher
from hyfi.joblib.batch.apply import decorator_apply
from hyfi.utils.logging import getLogger

logger = getLogger(__name__)


class PIPE:
    """
    A class to apply a pipe to a dataframe or a dictionary of dataframes.
    """

    @staticmethod
    def pipe(data, pipe):
        _func_ = pipe.get(SpecialKeys.FUNC)
        _fn = XC.partial(_func_)
        logger.info("Applying pipe: %s", _fn)
        if isinstance(data, dict):
            if "concat_dataframes" in str(_fn):
                return _fn(data, pipe)
            dfs = {}
            for df_no, df_name in enumerate(data):
                df_each = data[df_name]

                logger.info(
                    "Applying pipe to dataframe [%s], %d/%d",
                    df_name,
                    df_no + 1,
                    len(data),
                )

                pipe[SpecialKeys.SUFFIX.value] = df_name
                dfs[df_name] = _fn(df_each, pipe)
            return dfs
        return _fn(data, pipe)

    @staticmethod
    def apply(
        func,
        series,
        description=None,
        use_batcher=True,
        minibatch_size=None,
        num_workers=None,
        **kwargs,
    ):
        batcher_instance = batcher.batcher_instance
        if use_batcher and batcher_instance is not None:
            batcher_minibatch_size = batcher_instance.minibatch_size
            if minibatch_size is None:
                minibatch_size = batcher_minibatch_size
            if num_workers is not None:
                batcher_instance.procs = int(num_workers)
            if batcher_instance.procs > 1:
                batcher_instance.minibatch_size = min(
                    int(len(series) / batcher_instance.procs) + 1, minibatch_size
                )
                logger.info(
                    f"Using batcher with minibatch size: {batcher_instance.minibatch_size}"
                )
                results = decorator_apply(
                    func,
                    batcher_instance,
                    description=description,  # type: ignore
                )(series)
                if batcher_instance is not None:
                    batcher_instance.minibatch_size = batcher_minibatch_size
                return results

        if batcher_instance is None:
            logger.info("Warning: batcher not initialized")
        tqdm.pandas(desc=description)
        return series.progress_apply(func)



"""File I/O functions"""
import os
import re
from glob import glob
from pathlib import Path, PosixPath, WindowsPath
from typing import List, Union

from hyfi.utils.logging import getLogger

logger = getLogger(__name__)

PathLikeType = Union[str, PosixPath, WindowsPath]


def is_valid_regex(expr: str) -> bool:
    """Check if a string is a valid regular expression"""
    try:
        if expr.startswith("r:"):
            expr = expr[2:]
        else:
            return False
        re.compile(expr)
        return True
    except re.error:
        return False


def glob_re(
    pattern: str,
    base_dir: str,
    recursive: bool = False,
) -> list:
    """Glob files matching a regular expression"""
    if is_valid_regex(pattern):
        pattern = pattern[2:]
        rpattern = re.compile(pattern)  # type: ignore
        files = []
        if recursive:
            for dirpath, dirnames, filenames in os.walk(base_dir):
                files += [
                    os.path.join(dirpath, file)
                    for file in filenames
                    if rpattern.search(file)
                ]
        else:
            files = [
                os.path.join(base_dir, file)
                for file in os.listdir(base_dir)
                if rpattern.search(file)
            ]
    else:
        file = os.path.join(base_dir, pattern) if base_dir else pattern
        files = glob(file, recursive=recursive)
    return files


def get_filepaths(
    filename_patterns: Union[List[PathLikeType], PathLikeType],
    base_dir: Union[str, PosixPath, WindowsPath] = "",
    recursive: bool = True,
    verbose: bool = False,
    **kwargs,
) -> List[str]:
    """Get a list of filepaths from a list of filename patterns"""
    if filename_patterns is None:
        raise ValueError("filename_patterns must be specified")
    if isinstance(filename_patterns, (PosixPath, WindowsPath)):
        filename_patterns = str(filename_patterns)
    if isinstance(filename_patterns, str):
        filename_patterns = [filename_patterns]
    filepaths = []
    base_dir = str(base_dir) if base_dir else ""
    for file in filename_patterns:
        file = str(file)
        filepath = os.path.join(base_dir, file) if base_dir else file
        if os.path.exists(filepath):
            if Path(filepath).is_file():
                filepaths.append(filepath)
        else:
            if os.path.dirname(file) != "":
                _dir = os.path.dirname(file)
                file = os.path.basename(file)
                base_dir = os.path.join(base_dir, _dir) if base_dir else _dir
            filepaths += glob_re(file, base_dir, recursive=recursive)
    filepaths = [fp for fp in filepaths if Path(fp).is_file()]
    if verbose:
        logger.info(f"Processing [{len(filepaths)}] files from {filename_patterns}")

    return filepaths


def get_files_from_archive(archive_path: str, filetype: str = ""):
    """Get a list of files from an archive"""
    import tarfile
    from zipfile import ZipFile

    if ".tar.gz" in archive_path:
        logger.info(f"::Extracting files from {archive_path} with tar.gz")
        archive_handle = tarfile.open(archive_path, "r:gz")
        files = [
            (file, file.name) for file in archive_handle.getmembers() if file.isfile()
        ]
        open_func = archive_handle.extractfile
    elif ".tar.bz2" in archive_path:
        logger.info(f"::Extracting files from {archive_path} with tar.bz2")
        archive_handle = tarfile.open(archive_path, "r:bz2")
        files = [
            (file, file.name) for file in archive_handle.getmembers() if file.isfile()
        ]
        open_func = archive_handle.extractfile
    elif ".zip" in archive_path:
        logger.info(f"::Extracting files from {archive_path} with zip")
        archive_handle = ZipFile(archive_path)
        files = [
            (file, file.encode("cp437").decode("euc-kr"))
            for file in archive_handle.namelist()
        ]
        open_func = archive_handle.open
    else:
        # print(f'::{archive_path} is not archive, use generic method')
        files = [(archive_path, os.path.basename(archive_path))]
        archive_handle = None
        open_func = None
    if filetype:
        files = [file for file in files if filetype in file[1]]

    return files, archive_handle, open_func


def read(uri, mode="rb", encoding=None, head=None, **kwargs) -> bytes:
    """Read data from a file or url"""
    uri = str(uri)
    if uri.startswith("http"):
        import requests

        if mode == "r" and head is not None and isinstance(head, int):
            r = requests.get(uri, stream=True)
            r.raw.decode_content = True
            return r.raw.read(head)
        return requests.get(uri, **kwargs).content
    # elif uri.startswith("s3://"):
    #     import boto3

    #     s3 = boto3.resource("s3")
    #     bucket, key = uri.replace("s3://", "").split("/", 1)
    #     obj = s3.Object(bucket, key)
    #     return obj.get()["Body"].read()
    else:
        with open(uri, mode=mode, encoding=encoding) as f:
            if mode == "r" and head is not None and isinstance(head, int):
                return f.read(head)
            return f.read()


def is_file(a, *p) -> bool:
    """Check if path is a file"""
    _path = os.path.join(a, *p)
    return Path(_path).is_file()


def is_dir(a, *p) -> bool:
    """Check if path is a directory"""
    _path = os.path.join(a, *p)
    return Path(_path).is_dir()


def check_path(_path: str, alt_path: str = "") -> str:
    """Check if path exists, return alt_path if not"""
    return _path if os.path.exists(_path) else alt_path


def mkdir(_path: str) -> str:
    """Create directory if it does not exist"""
    if _path is None:
        return ""
    Path(_path).mkdir(parents=True, exist_ok=True)
    return _path


def exists(a, *p) -> bool:
    """Check if path exists"""
    if a is None:
        return False
    _path = os.path.join(a, *p)
    return os.path.exists(_path)


def join_path(a, *p) -> str:
    """Join path components intelligently."""
    if not p or p[0] is None:
        return a
    p = [str(_p) for _p in p]
    return os.path.join(*p) if a is None else os.path.join(a, *p)



import os
from typing import Union

import pandas as pd

from hyfi.utils.file import get_filepaths
from hyfi.utils.func import elapsed_timer
from hyfi.utils.logging import getLogger

logger = getLogger(__name__)


def is_dataframe(data) -> bool:
    """Check if data is a pandas dataframe"""
    return isinstance(data, pd.DataFrame)


def concat_data(
    data,
    columns=None,
    add_key_as_name: bool = False,
    name_column: str = "_name_",
    ignore_index: bool = True,
    **kwargs,
):
    """Concatenate dataframes"""
    if isinstance(data, dict):
        logger.info(f"Concatenating {len(data)} dataframes")
        dfs = []
        for df_name in data:
            df_each = data[df_name]
            if isinstance(columns, list):
                _columns = [c for c in columns if c in df_each.columns]
                df_each = df_each[_columns]
            if add_key_as_name:
                df_each[name_column] = df_name
            dfs.append(df_each)
        return pd.concat(dfs, ignore_index=ignore_index) if dfs else None
    elif isinstance(data, list):
        logger.info(f"Concatenating {len(data)} dataframes")
        return pd.concat(data, ignore_index=ignore_index) if len(data) > 0 else None
    else:
        logger.warning("Warning: data is not a dict")
        return data


def load_data(
    filename,
    base_dir=None,
    filetype=None,
    verbose=False,
    **kwargs,
):
    """Load data from a file or a list of files"""
    concatenate = kwargs.pop("concatenate", False)
    ignore_index = kwargs.pop("ignore_index", False)
    if filename is not None:
        filename = str(filename)

    if filename.startswith("http"):
        if not filetype:
            filetype = filename.split(".")[-1]
        if filetype not in ["csv", "parquet"]:
            raise ValueError("`file` should be a csv or a parquet file.")
        kwargs["filetype"] = filetype
        return load_dataframe(filename, verbose=verbose, **kwargs)

    if base_dir:
        filepaths = get_filepaths(filename, base_dir)
    else:
        filepaths = get_filepaths(filename)
    if verbose:
        logger.info(f"Loading {len(filepaths)} dataframes from {filepaths}")

    data = {
        os.path.basename(f): load_dataframe(
            f, verbose=verbose, filetype=filetype, **kwargs
        )
        for f in filepaths
    }
    data = {k: v for k, v in data.items() if v is not None}
    if len(data) == 1:
        return list(data.values())[0]
    elif len(filepaths) > 1:
        if concatenate:
            return pd.concat(data.values(), ignore_index=ignore_index)
        else:
            return data
    else:
        logger.warning(f"No files found for {filename}")
        return None


def load_dataframe(
    filename: str,
    base_dir: str = "",
    columns: list = None,  # type: ignore
    index_col=None,
    verbose: bool = False,
    **kwargs,
) -> Union[pd.DataFrame, None]:
    """Load a dataframe from a file"""
    dtype = kwargs.pop("dtype", None)
    if isinstance(dtype, list):
        dtype = {k: "str" for k in dtype}
    parse_dates = kwargs.pop("parse_dates", False)

    filetype = kwargs.pop("filetype", None) or "parquet"
    if filename.startswith("http"):
        filepath = filename
    else:
        fileinfo = os.path.splitext(filename)
        filename = fileinfo[0]
        filetype = fileinfo[1] if len(fileinfo) > 1 else filetype
        filetype = "." + filetype.replace(".", "")
        filename = f"{filename}{filetype}"
        filepath = os.path.join(base_dir, filename) if base_dir else filename
        if not os.path.exists(filepath):
            logger.warning(f"File {filepath} does not exist")
            return None
    if verbose:
        logger.info(f"Loading data from {filepath}")
    with elapsed_timer(format_time=True) as elapsed:
        if "csv" in filetype or "tsv" in filetype:
            delimiter = kwargs.pop("delimiter", "\t") if "tsv" in filetype else None
            data = pd.read_csv(
                filepath,
                index_col=index_col,
                dtype=dtype,
                parse_dates=parse_dates,
                delimiter=delimiter,
            )
        elif "parquet" in filetype:
            engine = kwargs.pop("engine", "pyarrow")
            if engine not in ["pyarrow", "fastparquet"]:
                engine = "auto"
            data = pd.read_parquet(filepath, engine=engine)  # type: ignore
        else:
            raise ValueError("filetype must be .csv or .parquet")
        if isinstance(columns, list):
            columns = [c for c in columns if c in data.columns]
            data = data[columns]
        if verbose:
            logger.info(" >> elapsed time to load data: %s", elapsed())
    return data


def save_data(
    data: Union[pd.DataFrame, dict],
    filename: str,
    base_dir: str = "",
    columns=None,
    index: bool = False,
    filetype="parquet",
    suffix: str = "",
    verbose: bool = False,
    **kwargs,
):
    """Save data to a file"""
    if filename is None:
        raise ValueError("filename must be specified")
    fileinfo = os.path.splitext(filename)
    filename = fileinfo[0]
    filetype = fileinfo[1] if len(fileinfo) > 1 else filetype
    filetype = "." + filetype.replace(".", "")
    if suffix:
        filename = f"{filename}-{suffix}{filetype}"
    else:
        filename = f"{filename}{filetype}"
    filepath = os.path.join(base_dir, filename) if base_dir else filename
    base_dir = os.path.dirname(filepath)
    filename = os.path.basename(filepath)
    os.makedirs(base_dir, exist_ok=True)

    if isinstance(data, dict):
        for k, v in data.items():
            save_data(
                v,
                filename,
                base_dir=base_dir,
                columns=columns,
                index=index,
                filetype=filetype,
                suffix=k,
                verbose=verbose,
                **kwargs,
            )
    elif is_dataframe(data):
        logger.info("Saving dataframe to %s", filepath)
        if isinstance(columns, list):
            columns = [c for c in columns if c in data.columns]
            data = data[columns]
        with elapsed_timer(format_time=True) as elapsed:
            if "csv" in filetype or "tsv" in filetype:
                data.to_csv(filepath, index=index)
            elif "parquet" in filetype:
                compression = kwargs.get("compression", "gzip")
                engine = kwargs.get("engine", "pyarrow")
                if engine not in ["pyarrow", "fastparquet"]:
                    engine = "auto"
                data.to_parquet(filepath, compression=compression, engine=engine)  # type: ignore
            else:
                raise ValueError("filetype must be .csv or .parquet")
            if verbose:
                logger.info(" >> elapsed time to save data: %s", elapsed())
    else:
        raise ValueError(f"Unsupported data type: {type(data)}")
